"""Exceptions used by hangups."""


class HangupsError(Exception):
    """An ambiguous error occurred."""


class NetworkError(HangupsError):
    """A network error occurred."""


class ConversationTypeError(HangupsError):
    """An action was performed on a conversation that doesn't support it."""



"""Google login authentication using OAuth 2.0.

Logging into Hangouts using OAuth2 requires a private scope only whitelisted
for certain clients. This module uses the client ID and secret from iOS, so it
will appear to Google to be an iOS device. Access can be revoked from this
page:
    https://security.google.com/settings/security/activity

This module should avoid logging any sensitive login information.

This module may be tested by invoking it directly:
    python -m hangups.auth
"""

import getpass
import logging
import os
import platform
import urllib.parse
import urllib.request

from datetime import datetime, timedelta

import aiohttp

try:
    from aiohttp_socks import ProxyConnector
except ImportError:
    ProxyConnector = None

from hangups import version

logger = logging.getLogger(__name__)
# Set the logging level for requests to at least INFO, since the DEBUG level
# will log sensitive data:
if logging.getLogger('requests').isEnabledFor(logging.DEBUG):
    logging.getLogger('requests').setLevel(logging.INFO)

OAUTH2_CLIENT_ID = '936475272427.apps.googleusercontent.com'
OAUTH2_CLIENT_SECRET = 'KWsJlkaMn1jGLxQpWxMnOox-'
OAUTH2_SCOPES = [
    'https://www.google.com/accounts/OAuthLogin',
    'https://www.googleapis.com/auth/userinfo.email',
]
# Note that '+' separating scopes must not be escaped by urlencode
OAUTH2_LOGIN_URL = (
    'https://accounts.google.com/o/oauth2/programmatic_auth?{}'.format(
        urllib.parse.urlencode(dict(
            scope='+'.join(OAUTH2_SCOPES),
            client_id=OAUTH2_CLIENT_ID,
            device_name='hangups',
        ), safe='+')
    )
)
OAUTH2_TOKEN_REQUEST_URL = 'https://accounts.google.com/o/oauth2/token'
FORM_SELECTOR = '#gaia_loginform'
EMAIL_SELECTOR = '#Email'
PASSWORD_SELECTOR = '#Passwd'
VERIFICATION_FORM_SELECTOR = '#challenge'
TOTP_CHALLENGE_SELECTOR = '[action="/signin/challenge/totp/2"]'
PHONE_CHALLENGE_SELECTOR = '[action="/signin/challenge/ipp/4"]'
TOTP_CODE_SELECTOR = '#totpPin'
PHONE_CODE_SELECTOR = '#idvPreregisteredPhonePin'
USER_AGENT = 'hangups/{} ({} {})'.format(
    version.__version__, platform.system(), platform.machine()
)

MANUAL_LOGIN_INSTRUCTIONS = '''
To sign in with your Google account:

    1) Open the URL provided below in your browser.

    2) Log into your Google account normally.

    3) You should be redirected to a loading screen. Copy the `oauth_code`
    cookie value set by this page and paste it here.

To obtain the cookie value using Chrome or Firefox:

    1) Press F12 to open developer tools.

    2) Select the "Application" (Chrome) or "Storage" (Firefox) tab.

    3) In the sidebar, expand "Cookies" and select
    `https://accounts.google.com`.

    4) In the cookie list, double click on the value for the `oauth_code`
    cookie to select it, and copy the value.

{}
'''.format(OAUTH2_LOGIN_URL)


class GoogleAuthError(Exception):
    """A Google authentication request failed."""


class CredentialsPrompt:
    """Callbacks for prompting user for their Google account credentials.

    This implementation prompts the user in a terminal using standard in/out.
    """

    @staticmethod
    async def get_authorization_code():
        """Prompt for authorization code.

        Returns:
            str: Google account authorization code.
        """
        print(MANUAL_LOGIN_INSTRUCTIONS)
        return input('Authorization code: ')


class RefreshTokenCache:
    """File-based cache for refresh token.

    Args:
        filename (str): Path to file where refresh token will be cached.
    """

    def __init__(self, filename):
        self._filename = filename

    async def get(self):
        """Get cached refresh token.

        Returns:
            Cached refresh token, or ``None`` on failure.
        """
        logger.info(
            'Loading refresh_token from %s', repr(self._filename)
        )
        try:
            with open(self._filename) as f:
                return f.read()
        except IOError as e:
            logger.info('Failed to load refresh_token: %s', e)

    async def set(self, refresh_token):
        """Cache a refresh token, ignoring any failure.

        Args:
            refresh_token (str): Refresh token to cache.
        """
        logger.info('Saving refresh_token to %s', repr(self._filename))
        try:
            with open(self._filename, 'w') as f:
                f.write(refresh_token)
        except IOError as e:
            logger.warning('Failed to save refresh_token: %s', e)


class TokenManager:
    def __init__(self, refresh_token_cache):
        connector = None
        try:
            http_proxy = urllib.request.getproxies()["http"]
        except KeyError:
            pass
        else:
            if ProxyConnector:
                # if the HTTP_PROXY environment variable is found set it up and assume
                # we're in a debugging environment so disable certificate verification
                # as well.
                logger.info("Found http_proxy environment, assuming debug "
                            "environment and disabling TLS certificate verification")
                connector = ProxyConnector.from_url(http_proxy, verify_ssl=False)
            else:
                logger.warning("http_proxy is set, but aiohttp-socks is not installed")

        # a requests.Session for handling our requests
        self.session = aiohttp.ClientSession(connector=connector, headers={
            "User-Agent": USER_AGENT,
        })

        # the refresh_token_cache so we can use and update it as necessary.
        self.refresh_token_cache = refresh_token_cache

        # access_token is the normal oauth access token which is used to get
        # the dynamite token. It's only used during run time and isn't stored
        # in the cache.
        self.access_token = None

        # oauth_expiration get's set to a datetime of when the oauth token
        # expires. This is necessary because a dynamite token expires hourly
        # but oauth tokens expire about every 24 hours, so we need to know to
        # refresh the oauth token so we can request a new dynamite token when
        # that happens.
        self.oauth_expiration = None

        # storage for the dynamite token and the datetime for when it will
        # expire.
        self.dynamite_token = None
        self.dynamite_expiration = None

    async def _token_request(self, data):
        """Make OAuth token request.

        Raises GoogleAuthError if authentication fails.

        Returns dict response.
        """
        try:
            r = await self.session.post(OAUTH2_TOKEN_REQUEST_URL, data=data)
            r.raise_for_status()
        except aiohttp.ClientError as e:
            raise GoogleAuthError('Token request failed: {}'.format(e))
        else:
            res = await r.json()
            # If an error occurred, a key 'error' will contain an error code.
            if 'error' in res:
                raise GoogleAuthError(
                    'Token request error: {!r}'.format(res['error'])
                )
            return res

    @staticmethod
    async def from_authorization_code(authorization_code, refresh_token_cache):
        r = TokenManager(refresh_token_cache)

        data = {
            'client_id': OAUTH2_CLIENT_ID,
            'client_secret': OAUTH2_CLIENT_SECRET,
            'code': authorization_code,
            'grant_type': 'authorization_code',
            'redirect_uri': 'urn:ietf:wg:oauth:2.0:oob',
        }

        res = await r._token_request(data)

        # store the new access token
        r.access_token = res['access_token']

        # set it's expiration
        expires_in = timedelta(seconds=int(res['expires_in']))
        r.oauth_expiration = datetime.now() + expires_in

        # store the new refresh token
        await r.refresh_token_cache.set(res['refresh_token'])

        # finally get the dynamite token
        await r._refresh_dynamite()

        return r

    async def _refresh_oauth(self):
        """Request a new access token from the refresh_token stored in the
           cache."""

        refresh_token = await self.refresh_token_cache.get()
        if refresh_token is None:
            raise GoogleAuthError("Refresh token not found")

        data = {
            'client_id': OAUTH2_CLIENT_ID,
            'client_secret': OAUTH2_CLIENT_SECRET,
            'grant_type': 'refresh_token',
            'refresh_token': refresh_token,
        }

        res = await self._token_request(data)

        # store the new access token
        self.access_token = res['access_token']

        # store the new expiration time
        expires_in = timedelta(seconds=int(res['expires_in']))
        self.oauth_expiration = datetime.now() + expires_in

    @staticmethod
    async def from_refresh_token(refresh_token_cache):
        r = TokenManager(refresh_token_cache)

        await r._refresh_dynamite()

        return r

    async def _refresh_dynamite(self):
        """Use the oauth access token to get a dynamite token.

        Raises GoogleAuthError if the dynamite token couldn't be acquired.
        """

        # check the oauth_expiration before using the access token
        if self.oauth_expiration is None or datetime.now() >= self.oauth_expiration:
            await self._refresh_oauth()

        headers = {
            'Authorization': 'Bearer {}'.format(self.access_token),
            'Content-Type': 'application/x-www-form-urlencoded',
        }
        data = {
            'app_id': 'com.google.Dynamite',
            'client_id': '576267593750-sbi1m7khesgfh1e0f2nv5vqlfa4qr72m.apps.googleusercontent.com',
            'passcode_present': 'YES',
            'response_type': 'token',
            'scope': ' '.join([
                'https://www.googleapis.com/auth/dynamite',
                'https://www.googleapis.com/auth/drive',
                'https://www.googleapis.com/auth/mobiledevicemanagement',
                'https://www.googleapis.com/auth/notifications',
                'https://www.googleapis.com/auth/supportcontent',
                'https://www.googleapis.com/auth/chat.integration',
                'https://www.googleapis.com/auth/peopleapi.readonly',
            ])
        }

        try:
            r = await self.session.post('https://oauthaccountmanager.googleapis.com/v1/issuetoken',
                                        headers=headers, data=data)
            r.raise_for_status()
        except aiohttp.ClientError as e:
            raise GoogleAuthError('OAuthLogin request failed: {}'.format(e))

        body = await r.json()

        try:
            self.dynamite_token = body['token']
            expires_in = timedelta(seconds=int(body['expiresIn']))
            self.dynamite_expiration = datetime.now() + expires_in
        except IndexError:
            raise GoogleAuthError('Failed to find the dynamite token')

    async def get(self):
        if self.dynamite_expiration is None or datetime.now() >= self.dynamite_expiration:
            await self._refresh_dynamite()

        return self.dynamite_token


async def get_auth(credentials_prompt, refresh_token_cache):
    """Authenticate with Google.

    Args:
        refresh_token_cache (RefreshTokenCache): Cache to use so subsequent
            logins may not require credentials.
        credentials_prompt (CredentialsPrompt): Prompt to use if credentials
            are required to log in.

    Returns:
        dict: Google session cookies.

    Raises:
        GoogleAuthError: If authentication with Google fails.
    """

    try:
        logger.info('Authenticating with refresh token')
        token_manager = await TokenManager.from_refresh_token(refresh_token_cache)
    except GoogleAuthError as e:
        logger.info('Failed to authenticate using refresh token: %s', e)
        logger.info('Authenticating with credentials')
        authorization_code = await credentials_prompt.get_authorization_code()
        token_manager = await TokenManager.from_authorization_code(
            authorization_code, refresh_token_cache
        )

    logger.info('Authentication successful')

    return token_manager


def get_auth_stdin(refresh_token_filename):
    """Simple wrapper for :func:`get_auth` that prompts the user using stdin.

    Args:
        refresh_token_filename (str): Path to file where refresh token will be
            cached.

    Raises:
        GoogleAuthError: If authentication with Google fails.
    """
    refresh_token_cache = RefreshTokenCache(refresh_token_filename)
    return get_auth(
        CredentialsPrompt(), refresh_token_cache
    )


if __name__ == '__main__':
    logging.basicConfig(level=logging.DEBUG)
    print(get_auth_stdin('refresh_token.txt'))



"""ConversationEvent base class and subclasses.

These classes are wrappers for hangouts_pb2.Event instances. Parsing is done
through property methods, which prefer logging warnings to raising exceptions.
"""

import logging
from urllib.parse import urljoin, urlencode

from hangups import parsers, message_parser, user, hangouts_pb2

logger = logging.getLogger(__name__)
chat_message_parser = message_parser.ChatMessageParser()


class ConversationEvent:
    """An event which becomes part of the permanent record of a conversation.

    This is a wrapper for the ``Event`` message, which may contain one of many
    subtypes, represented here as other subclasses.

    Args:
        event: ``Event`` message.
    """

    def __init__(self, event):
        self._event = event  # Event
        self._message = self._event.body.message_posted.message

    @property
    def timestamp(self):
        """When the event occurred (:class:`datetime.datetime`)."""
        return parsers.from_timestamp(self._message.create_time)

    @property
    def user_id(self):
        """Who created the event (:class:`~hangups.user.UserID`)."""

        return self._message.creator.user_id.id

    @property
    def conversation_id(self):
        """ID of the conversation containing the event (:class:`str`)."""

        if self._event.group_id.HasField('dm_id'):
            return f'dm:{self._event.group_id.dm_id.dm_id}'
        else:
            return f'space:{self._event.group_id.space_id.space_id}'

    @property
    def id_(self):
        """ID of this event (:class:`str`)."""
        if self._event.HasField('user_revision'):
            return self._event.user_revision.timestamp

        return self._event.group_revision.timestamp

class ChatMessageSegment:
    """A segment of a chat message in :class:`ChatMessageEvent`.

    Args:
        text (str): Text of the segment.
        segment_type: (optional) One of ``SEGMENT_TYPE_TEXT``,
            ``SEGMENT_TYPE_LINE_BREAK``, or ``SEGMENT_TYPE_LINK``. Defaults to
            ``SEGMENT_TYPE_TEXT``, or ``SEGMENT_TYPE_LINK`` if ``link_target``
            is specified.
        is_bold (bool): (optional) Whether the text is bold. Defaults to
            ``False``.
        is_italic (bool): (optional) Whether the text is italic. Defaults to
            ``False``.
        is_strikethrough (bool): (optional) Whether the text is struck through.
            Defaults to ``False``.
        is_underline (bool): (optional) Whether the text is underlined.
            Defaults to ``False``.
        link_target (str): (option) URL to link to. Defaults to ``None``.
    """

    def __init__(self, text, segment_type=None,
                 is_bold=False, is_italic=False, is_strikethrough=False,
                 is_underline=False, link_target=None):
        """Create a new chat message segment."""
        if segment_type is not None:
            self.type_ = segment_type
        elif link_target is not None:
            self.type_ = hangouts_pb2.SEGMENT_TYPE_LINK
        else:
            self.type_ = hangouts_pb2.SEGMENT_TYPE_TEXT
        self.text = text
        self.is_bold = is_bold
        self.is_italic = is_italic
        self.is_strikethrough = is_strikethrough
        self.is_underline = is_underline
        self.link_target = link_target

    @staticmethod
    def from_str(text):
        """Construct :class:`ChatMessageSegment` list parsed from a string.

        Args:
            text (str): Text to parse. May contain line breaks, URLs and
                formatting markup (simplified Markdown and HTML) to be
                converted into equivalent segments.

        Returns:
            List of :class:`ChatMessageSegment` objects.
        """
        segment_list = chat_message_parser.parse(text)
        return [ChatMessageSegment(segment.text, **segment.params)
                for segment in segment_list]

    @staticmethod
    def deserialize(segment):
        """Construct :class:`ChatMessageSegment` from ``Segment`` message.

        Args:
            segment: ``Segment`` message to parse.

        Returns:
            :class:`ChatMessageSegment` object.
        """
        link_target = segment.link_data.link_target
        return ChatMessageSegment(
            segment.text, segment_type=segment.type,
            is_bold=segment.formatting.bold,
            is_italic=segment.formatting.italic,
            is_strikethrough=segment.formatting.strikethrough,
            is_underline=segment.formatting.underline,
            link_target=None if link_target == '' else link_target
        )

    def serialize(self):
        """Serialize this segment to a ``Segment`` message.

        Returns:
            ``Segment`` message.
        """
        segment = hangouts_pb2.Segment(
            type=self.type_,
            text=self.text,
            formatting=hangouts_pb2.Formatting(
                bold=self.is_bold,
                italic=self.is_italic,
                strikethrough=self.is_strikethrough,
                underline=self.is_underline,
            ),
        )
        if self.link_target is not None:
            segment.link_data.link_target = self.link_target
        return segment


class ChatMessageEvent(ConversationEvent):
    """An event that adds a new message to a conversation.

    Corresponds to the ``MessageEvent`` message.
    """

    @property
    def text(self):
        """Text of the message without formatting (:class:`str`)."""

        return self._event.body.message_posted.message.text_body

    @property
    def segments(self):
        """List of :class:`ChatMessageSegment` in message (:class:`list`)."""
        seg_list = self._event.chat_message.message_content.segment
        return [ChatMessageSegment.deserialize(seg) for seg in seg_list]

    @property
    def attachments(self):
        """List of attachments in the message (:class:`list`)."""
        raw_annotations = self._event.body.message_posted.message.annotations
        if raw_annotations is None:
            raw_annotations = []

        annotations = []
        for annotation in raw_annotations:
            # instead of bothering with checking the type, we just look for the
            # fields that have image urls.

            if annotation.HasField('upload_metadata'):
                # Uploads have an attachment token that helps control access to
                # the image data.

                base = 'https://chat.google.com/api/get_attachment_url'

                params = {
                    'url_type': 'FIFE_URL',
                    'attachment_token': annotation.upload_metadata.attachment_token,
                }
                query = '?' + urlencode(params)

                annotations.append(urljoin(base, query))

            elif annotation.HasField('url_metadata'):
                if annotation.url_metadata.HasField('image_url'):
                    annotations.append(annotation.url_metadata.image_url)

        return annotations

    @property
    def local_id(self):
        return self._message.local_id

    @property
    def msg_id(self):
        """ID of this message (:class:`str`)."""
        return self._message.id.message_id

    @property
    def parent_msg_id(self):
        """ID of this message (:class:`str`)."""
        return self._message.id.parent_id.topic_id.topic_id


class OTREvent(ConversationEvent):
    """An event that changes a conversation's OTR (history) mode.

    Corresponds to the ``OTRModification`` message.
    """

    @property
    def new_otr_status(self):
        """The conversation's new OTR status.

        May be either ``OFF_THE_RECORD_STATUS_OFF_THE_RECORD`` or
        ``OFF_THE_RECORD_STATUS_ON_THE_RECORD``.
        """
        return self._event.otr_modification.new_otr_status

    @property
    def old_otr_status(self):
        """The conversation's old OTR status.

        May be either ``OFF_THE_RECORD_STATUS_OFF_THE_RECORD`` or
        ``OFF_THE_RECORD_STATUS_ON_THE_RECORD``.
        """
        return self._event.otr_modification.old_otr_status


class RenameEvent(ConversationEvent):
    """An event that renames a conversation.

    Corresponds to the ``ConversationRename`` message.
    """

    @property
    def new_name(self):
        """The conversation's new name (:class:`str`).

        May be an empty string if the conversation's name was cleared.
        """
        return self._event.conversation_rename.new_name

    @property
    def old_name(self):
        """The conversation's old name (:class:`str`).

        May be an empty string if the conversation had no previous name.
        """
        return self._event.conversation_rename.old_name


class MembershipChangeEvent(ConversationEvent):
    """An event that adds or removes a conversation participant.

    Corresponds to the ``MembershipChange`` message.
    """

    @property
    def type_(self):
        """The type of membership change.

        May be either ``MEMBERSHIP_CHANGE_TYPE_JOIN`` or
        ``MEMBERSHIP_CHANGE_TYPE_LEAVE``.
        """
        return self._event.membership_change.type

    @property
    def participant_ids(self):
        """:class:`~hangups.user.UserID` of users involved (:class:`list`)."""
        return [user.UserID(chat_id=id_.chat_id, gaia_id=id_.gaia_id)
                for id_ in self._event.membership_change.participant_ids]


class HangoutEvent(ConversationEvent):
    """An event that is related to a Hangout voice or video call.

    Corresponds to the ``HangoutEvent`` message.
    """

    @property
    def event_type(self):
        """The Hangout event type.

        May be one of ``HANGOUT_EVENT_TYPE_START``, ``HANGOUT_EVENT_TYPE_END``,
        ``HANGOUT_EVENT_TYPE_JOIN``, ``HANGOUT_EVENT_TYPE_LEAVE``,
        ``HANGOUT_EVENT_TYPE_COMING_SOON``, or ``HANGOUT_EVENT_TYPE_ONGOING``.
        """
        return self._event.hangout_event.event_type


class GroupLinkSharingModificationEvent(ConversationEvent):
    """An event that modifies a conversation's group link sharing status.

    Corresponds to the ``GroupLinkSharingModification`` message.
    """

    @property
    def new_status(self):
        """The new group link sharing status.

        May be either ``GROUP_LINK_SHARING_STATUS_ON`` or
        ``GROUP_LINK_SHARING_STATUS_OFF``.
        """
        return self._event.group_link_sharing_modification.new_status



"""Tests for the user module"""


import hangups.user
import hangups.hangouts_pb2


USER_ID = hangups.user.UserID(1, 1)


def test_default_type_detection_empty_0():
    # missing names
    user = hangups.user.User(
        USER_ID,
        full_name='',
        first_name='',
        photo_url='',
        canonical_email='',
        emails=[],
        is_self=False,
    )

    assert user.full_name == hangups.user.DEFAULT_NAME
    assert user.first_name == hangups.user.DEFAULT_NAME
    assert user.name_type == hangups.user.NameType.DEFAULT


def test_default_type_detection_empty_1():
    # missing names
    user = hangups.user.User(
        USER_ID,
        full_name=None,
        first_name=None,
        photo_url='',
        canonical_email='',
        emails=[],
        is_self=False,
    )

    assert user.full_name == hangups.user.DEFAULT_NAME
    assert user.first_name == hangups.user.DEFAULT_NAME
    assert user.name_type == hangups.user.NameType.DEFAULT


def test_default_type_detection_from_conv_part_data():
    # default user in 201904
    conv_part_data = hangups.hangouts_pb2.ConversationParticipantData(
        id=hangups.hangouts_pb2.ParticipantId(
            chat_id='1',
            gaia_id='1'
        ),
        fallback_name='unknown',
        invitation_status=hangups.hangouts_pb2.INVITATION_STATUS_ACCEPTED,
        participant_type=hangups.hangouts_pb2.PARTICIPANT_TYPE_GAIA,
        new_invitation_status=hangups.hangouts_pb2.INVITATION_STATUS_ACCEPTED,
    )

    user = hangups.user.User.from_conv_part_data(
        conv_part_data=conv_part_data,
        self_user_id=USER_ID
    )

    assert user.full_name == hangups.user.DEFAULT_NAME
    assert user.first_name == hangups.user.DEFAULT_NAME
    assert user.name_type == hangups.user.NameType.DEFAULT


def test_real_type():
    # regular name
    user = hangups.user.User(
        USER_ID,
        full_name='Joe Doe',
        first_name='Joe',
        photo_url='',
        canonical_email='',
        emails=[],
        is_self=False,
    )

    assert user.full_name == 'Joe Doe'
    assert user.first_name == 'Joe'
    assert user.name_type == hangups.user.NameType.REAL


def test_real_type_from_conv_part_data():
    conv_part_data = hangups.hangouts_pb2.ConversationParticipantData(
        id=hangups.hangouts_pb2.ParticipantId(
            chat_id='1',
            gaia_id='1'
        ),
        fallback_name='Joe Doe',
        invitation_status=hangups.hangouts_pb2.INVITATION_STATUS_ACCEPTED,
        participant_type=hangups.hangouts_pb2.PARTICIPANT_TYPE_GAIA,
        new_invitation_status=hangups.hangouts_pb2.INVITATION_STATUS_ACCEPTED,
    )

    user = hangups.user.User.from_conv_part_data(
        conv_part_data=conv_part_data,
        self_user_id=USER_ID
    )

    assert user.full_name == 'Joe Doe'
    assert user.first_name == 'Joe'
    assert user.name_type == hangups.user.NameType.REAL
